<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * @package qtype_kprime
 * @author Amr Hourani amr.hourani@id.ethz.ch
 * @copyright ETHz 2016 amr.hourani@id.ethz.ch
 */
defined('MOODLE_INTERNAL') || die();


/**
 * Restore plugin class that provides the necessary information
 * needed to restore one kprime qtype plugin.
 */
class restore_qtype_kprime_plugin extends restore_qtype_plugin {

    /**
     * Returns the paths to be handled by the plugin at question level.
     */
    protected function define_question_plugin_structure() {
        $result = array();

        // We used get_recommended_name() so this works.
        $elename = 'kprime';
        $elepath = $this->get_pathfor('/kprime');
        $result[] = new restore_path_element($elename, $elepath);

        // We used get_recommended_name() so this works.
        $elename = 'column';
        $elepath = $this->get_pathfor('/columns/column');
        $result[] = new restore_path_element($elename, $elepath);

        // We used get_recommended_name() so this works.
        $elename = 'row';
        $elepath = $this->get_pathfor('/rows/row');
        $result[] = new restore_path_element($elename, $elepath);

        // We used get_recommended_name() so this works.
        $elename = 'weight';
        $elepath = $this->get_pathfor('/weights/weight');
        $result[] = new restore_path_element($elename, $elepath);

        return $result;
    }

    /**
     * Process the qtype/multichoice element.
     */
    public function process_kprime($data) {
        global $DB;

        $data = (object) $data;
        $oldid = $data->id;

        // Detect if the question is created or mapped.
        $oldquestionid = $this->get_old_parentid('question');
        $newquestionid = $this->get_new_parentid('question');

        $questioncreated = (bool) $this->get_mappingid('question_created', $oldquestionid);

        // If the question has been created by restore, we need to create its
        // qtype_kprime_options too.
        if ($questioncreated) {
            // Adjust some columns.
            $data->questionid = $newquestionid;
            // Insert record.
            $newitemid = $DB->insert_record('qtype_kprime_options', $data);
            // Create mapping (needed for decoding links).
            $this->set_mapping('qtype_kprime_options', $oldid, $newitemid);
        }
    }

    /**
     * Detect if the question is created or mapped.
     *
     * @return bool
     */
    protected function is_question_created() {
        $oldquestionid = $this->get_old_parentid('question');
        $questioncreated = (bool) $this->get_mappingid('question_created', $oldquestionid);

        return $questioncreated;
    }

    /**
     * Process the qtype/kprime/columns/column.
     */
    public function process_column($data) {
        global $DB;

        if (!$this->is_question_created()) {
            return;
        }

        $data = (object) $data;
        $oldid = $data->id;

        $data->questionid = $this->get_new_parentid('question');
        $newitemid = $DB->insert_record('qtype_kprime_columns', $data);
        $this->set_mapping('qtype_kprime_columns', $oldid, $newitemid);
    }

    /**
     * Process the qtype/kprime/rows/row element.
     */
    public function process_row($data) {
        global $DB;

        if (!$this->is_question_created()) {
            return;
        }

        $data = (object) $data;
        $oldid = $data->id;

        $data->questionid = $this->get_new_parentid('question');
        $newitemid = $DB->insert_record('qtype_kprime_rows', $data);

        $this->set_mapping('qtype_kprime_rows', $oldid, $newitemid);
    }

    /**
     * Process the qtype/kprime/weights/weight element.
     */
    public function process_weight($data) {
        global $DB;

        if (!$this->is_question_created()) {
            return;
        }

        $data = (object) $data;
        $oldid = $data->id;

        $data->questionid = $this->get_new_parentid('question');
        $newitemid = $DB->insert_record('qtype_kprime_weights', $data);
        $this->set_mapping('qtype_kprime_weights', $oldid, $newitemid);
    }

    public function recode_response($questionid, $sequencenumber, array $response) {
        if (array_key_exists('_order', $response)) {
            $response['_order'] = $this->recode_option_order($response['_order']);
        }

        return $response;
    }

    /**
     * Recode the option order as stored in the response.
     *
     * @param string $order the original order.
     *
     * @return string the recoded order.
     */
    protected function recode_option_order($order) {
        $neworder = array();
        foreach (explode(',', $order) as $id) {
            if ($newid = $this->get_mappingid('qtype_kprime_rows', $id)) {
                $neworder[] = $newid;
            } else {
                $neworder[] = $id;
            }
        }

        return implode(',', $neworder);
    }

    /**
     * Return the contents of this qtype to be processed by the links decoder.
     */
    public static function define_decode_contents() {
        $contents = array();

        $fields = array('optiontext', 'optionfeedback'
        );
        $contents[] = new restore_decode_content('qtype_kprime_rows', $fields, 'qtype_kprime_rows');

        return $contents;
    }
}
